/*
  Copyright (c) 2009 Mikko Juola

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/


#include <iostream>
#include <fstream>
#include <string>
#include <map>
#include <vector>
#include <cstdlib>
#include "srandom.hpp"

#define program_name "seed_by_discoveries"

using namespace std;

vector<unsigned short> NameIndices;
vector<unsigned short> AppearanceIndices;
vector<short> NamesToAppearances;

class Object
{
    public:
        string Name;         // Name of the actual object (e.g. "identify")
        string Appearance;   // "wooden/granite/ZELGO MER, etc.
        int Index;           // The index that NetHack uses internally
        int AppearanceIndex;
        int NameIndex;       
        
        Object(string n, string a, int i) : Name(n), Appearance(a), Index(i)
        { AppearanceIndex = -1; NameIndex = -1; };
};

Object RingList[] = { Object("adornment", "wooden", 150),
                      Object("gain strength", "granite", 151),
                      Object("gain constitution", "opal", 152),
                      Object("increase accuracy", "clay", 153),
                      Object("increase damage", "coral", 154),
                      Object("protection", "black onyx", 155),
                      Object("regeneration", "moonstone", 156),
                      Object("searching", "tiger eye", 157),
                      Object("stealth", "jade", 158),
                      Object("sustain ability", "bronze", 159),
                      Object("levitation", "agate", 160),
                      Object("hunger", "topaz", 161),
                      Object("aggravate monster", "sapphire", 162),
                      Object("conflict", "ruby", 163),
                      Object("warning", "diamond", 164),
                      Object("poison resistance", "pearl", 165),
                      Object("fire resistance", "iron", 166),
                      Object("cold resistance", "brass", 167),
                      Object("shock resistance", "copper", 168),
                      Object("free action", "twisted", 169),
                      Object("slow digestion", "steel", 170),
                      Object("teleportation", "silver", 171),
                      Object("teleport control", "gold", 172),
                      Object("polymorph", "ivory", 173),
                      Object("polymorph control", "emerald", 174),
                      Object("invisibility", "wire", 175),
                      Object("see invisible", "engagement", 176),
                      Object("protection from shape changers", "shiny", 177) 
                    };
const int RingListSize = 28;

Object AmuletList[] = { Object("amulet of ESP", "circular", 178),
                        Object("amulet of life saving", "spherical", 179),
                        Object("amulet of strangulation", "oval", 180),
                        Object("amulet of restful sleep", "triangular", 181),
                        Object("amulet versus poison", "pyramidal", 182),
                        Object("amulet of change", "square", 183),
                        Object("amulet of unchanging",  "concave", 184),
                        Object("amulet of reflection", "hexagonal", 185),
                        Object("amulet of magical breathing", "octagonal", 186)
                      };
const int AmuletListSize = 9;

Object PotionList[] = { Object("gain ability", "ruby", 268),
                        Object("restore ability", "pink", 269),
                        Object("confusion", "orange", 270),
                        Object("blindness", "yellow", 271),
                        Object("paralysis", "emerald", 272),
                        Object("speed", "dark green", 273),
                        Object("levitation", "cyan", 274),
                        Object("hallucination", "sky blue", 275),
                        Object("invisibility", "brilliant blue", 276),
                        Object("see invisible", "magenta", 277),
                        Object("healing", "purple-red", 278),
                        Object("extra healing", "puce", 279),
                        Object("gain level", "milky", 280),
                        Object("enlightenment", "swirly", 281),
                        Object("monster detection", "bubbly", 282),
                        Object("object detection", "smoky", 283),
                        Object("gain energy", "cloudy", 284),
                        Object("sleeping", "effervescent", 285),
                        Object("full healing", "black", 286),
                        Object("polymorph", "golden", 287),
                        Object("booze", "brown", 288),
                        Object("sickness", "fizzy", 289),
                        Object("fruit juice", "dark", 290),
                        Object("acid", "white", 291),
                        Object("oil", "murky", 292)
                      };
const int PotionListSize = 25;

Object ScrollList[] = { Object("enchant armor", "ZELGO MER", 294),
                        Object("destroy armor", "JUYED AWK YACC", 295),
                        Object("confuse monster", "NR 9", 296),
                        Object("scare monster", "XIXAXA XOXAXA XUXAXA", 297),
                        Object("remove curse", "PRATYAVAYAH", 298),
                        Object("enchant weapon", "DAIYEN FOOELS", 299),
                        Object("create monster", "LEP GEX VEN ZEA", 300),
                        Object("taming", "PRIRUTSENIE", 301),
                        Object("genocide", "ELBIB YLOH", 302),
                        Object("light", "VERR YED HORRE", 303),
                        Object("teleportation", "VENZAR BORGAVVE", 304),
                        Object("gold detection", "THARR", 305),
                        Object("food detection", "YUM YUM", 306),
                        Object("identify", "KERNOD WEL", 307),
                        Object("magic mapping", "ELAM EBOW", 308),
                        Object("amnesia", "DUAM XNAHT", 309),
                        Object("fire", "ANDOVA BEGARIN", 310),
                        Object("earth", "KIRJE", 311),
                        Object("punishment", "VE FORBRYDERNE", 312),
                        Object("charging", "HACKEM MUCHE", 313),
                        Object("stinking cloud", "VELOX NEB", 314),
                        Object("(unused1)", "FOOBIE BLETCH", 315),
                        Object("(unused2)", "TEMOV", 316),
                        Object("(unused3)", "GARVEN DEH", 317),
                        Object("(unused4)", "READ ME", 318)
                      };
const int ScrollListSize = 25;

Object SpellbookList[] = { Object("dig", "parchment", 321),
                           Object("magic missile", "vellum", 322),
                           Object("fireball", "ragged", 323),
                           Object("cone of cold", "dog eared", 324),
                           Object("sleep", "mottled", 325),
                           Object("finger of death", "stained", 326),
                           Object("light", "cloth", 327),
                           Object("detect monsters", "leather", 328),
                           Object("healing", "white", 329),
                           Object("knock", "pink", 330),
                           Object("force bolt", "red", 331),
                           Object("confuse monster", "orange", 332),
                           Object("cure blindness", "yellow", 333),
                           Object("drain life", "velvet", 334),
                           Object("slow monster", "light green", 335),
                           Object("wizard lock", "dark green", 336),
                           Object("create monster", "turquoise", 337),
                           Object("detect food", "cyan", 338),
                           Object("cause fear", "light blue", 339),
                           Object("clairvoyance", "dark blue", 340),
                           Object("cure sickness", "indigo", 341),
                           Object("charm monster", "magenta", 342),
                           Object("haste self", "purple", 343),
                           Object("detect unseen", "violet", 344),
                           Object("levitation", "tan", 345),
                           Object("extra healing", "plaid", 346),
                           Object("restore ability", "light brown", 347),
                           Object("invisibility", "dark brown", 348),
                           Object("detect treasure", "gray", 349),
                           Object("remove curse", "wrinkled", 350),
                           Object("magic mapping", "dusty", 351),
                           Object("identify", "bronze", 352),
                           Object("turn undead", "copper", 353),
                           Object("polymorph", "silver", 354),
                           Object("teleport away", "gold", 355),
                           Object("create familiar", "glittering", 356),
                           Object("cancellation", "shining", 357),
                           Object("protection", "dull", 358),
                           Object("jumping", "thin", 359),
                           Object("stone to flesh", "thick", 360)
                       };
const int SpellbookListSize = 40;

Object WandList[] = { Object("light", "glass", 363),
                      Object("secret door detection", "balsa", 364),
                      Object("enlightenment", "crystal", 365),
                      Object("create monster", "maple", 366),
                      Object("wishing", "pine", 367),
                      Object("nothing", "oak", 368),
                      Object("striking", "ebony", 369),
                      Object("make invisible", "marble", 370),
                      Object("slow monster", "tin", 371),
                      Object("speed monster", "brass", 372),
                      Object("undead turning", "copper", 373),
                      Object("polymorph", "silver", 374),
                      Object("cancellation", "platinum", 375),
                      Object("teleportation", "iridium", 376),
                      Object("opening", "zinc", 377),
                      Object("locking", "aluminum", 378),
                      Object("probing", "uranium", 379),
                      Object("digging", "iron", 380),
                      Object("magic missile", "steel", 381),
                      Object("fire", "hexagonal", 382),
                      Object("cold", "short", 383),
                      Object("sleep", "runed", 384),
                      Object("death", "long", 385),
                      Object("lightning", "curved", 386),
                      Object("(unused1)", "forked", 387),
                      Object("(unused2)", "spiked", 388),
                      Object("(unused3)", "jeweled", 389) 
                    };
const int WandListSize = 27;

vector<Object> RingVector, AmuletVector, PotionVector, ScrollVector, SpellbookVector, WandVector;

int getVectorNameIndex(const vector<Object> &object_vector, const string &s)
{
    vector<Object>::const_iterator i1;
    for (i1 = object_vector.begin(); i1 != object_vector.end(); i1++)
        if (i1->Name == s)
            return i1->NameIndex;
    
    return -1;
};

int getVectorAppearanceIndex(const vector<Object> &object_vector, const string &s)
{
    vector<Object>::const_iterator i1;
    for (i1 = object_vector.begin(); i1 != object_vector.end(); i1++)
        if (i1->Appearance == s)
            return i1->AppearanceIndex;
    
    return -1;
};

string getVectorAppearance(const vector<Object> &object_vector, short index)
{
    vector<Object>::const_iterator i1;
    for (i1 = object_vector.begin(); i1 != object_vector.end(); i1++)
        if (i1->AppearanceIndex == index)
            return i1->Appearance;
    
    return "";
}

vector<Object> buildObjectVector(Object* object_list, int list_size)
{
    vector<Object> result;
    int i1;
    for (i1 = 0; i1 < list_size; i1++)
        result.push_back(object_list[i1]);

    return result;
};

void buildAllObjectVectors()
{
    RingVector = buildObjectVector(RingList, RingListSize);
    AmuletVector = buildObjectVector(AmuletList, AmuletListSize);
    PotionVector = buildObjectVector(PotionList, PotionListSize);
    ScrollVector = buildObjectVector(ScrollList, ScrollListSize);
    SpellbookVector = buildObjectVector(SpellbookList, SpellbookListSize);
    WandVector = buildObjectVector(WandList, WandListSize);
};

void buildIndices(Object* object_list, int list_size)
{
    int i1;
    for (i1 = 0; i1 < list_size; i1++)
    {
        object_list[i1].AppearanceIndex = AppearanceIndices.size();
        object_list[i1].NameIndex = NameIndices.size();
        AppearanceIndices.push_back(AppearanceIndices.size());
        NameIndices.push_back(NameIndices.size());
    }
};

void buildAllIndices()
{
    buildIndices(RingList, RingListSize);
    buildIndices(AmuletList, AmuletListSize);
    buildIndices(PotionList, PotionListSize);
    buildIndices(ScrollList, ScrollListSize);
    buildIndices(SpellbookList, SpellbookListSize);
    buildIndices(WandList, WandListSize);
        
    vector<unsigned short> copy = AppearanceIndices;
    AppearanceIndices = copy;
    copy = NameIndices;
    NameIndices = copy;
    
    NamesToAppearances.resize(AppearanceIndices.size());
    unsigned int i1;
    for (i1 = 0; i1 < AppearanceIndices.size(); i1++)
        NamesToAppearances[NameIndices[i1]] = AppearanceIndices[i1];
};

void fixNamesToAppearances(vector<short> &names_to_appearances, int first_index, int last_index)
{
    int i1;
    for (i1 = first_index; i1 < last_index; i1++)
        names_to_appearances[i1] = NamesToAppearances[i1];
}

void fixNamesToAppearances(vector<short> &names_to_appearances, const vector<Object> &object_vector)
{
    fixNamesToAppearances(names_to_appearances, object_vector[0].NameIndex, object_vector[0].NameIndex + object_vector.size());
}

bool shuffle(const vector<Object> &object_vector, 
             vector<short> &names_to_appearances, 
             const vector<short> &target_mapping,
             RandomState &rs)
{
    const int vector_size = object_vector.size();
    const int first_index = object_vector[0].NameIndex;
    const int last_index = first_index + vector_size;

    int i1;
    for (i1 = first_index; i1 < last_index; i1++)
    {
        int swap_target = i1 + (rs.random() % (vector_size - i1 + first_index));
        int temp = names_to_appearances[i1];
        names_to_appearances[i1] = names_to_appearances[swap_target];
        names_to_appearances[swap_target] = temp;
        if (target_mapping[i1] != -1 && target_mapping[i1] != names_to_appearances[i1])
            return false;
    };

    return true;
};

void showAppearances(const vector<Object> &object_vector, const vector<short> &names_to_appearances)
{
    vector<Object>::const_iterator i;
    for (i = object_vector.begin(); i != object_vector.end(); ++i)
    {
        string n = (*i).Name;
        string a = getVectorAppearance(object_vector, names_to_appearances[(*i).NameIndex]);
	cout << "  " << n << " = " << a << endl;
    }
}

void showAllAppearances(int seed, int offset)
{
    RandomState rs;
    rs.srandom(seed);
    for (int i = 0; i < offset; ++i)
        rs.random();

    vector<short> names_to_appearances(NamesToAppearances);
    vector<short> minus_one_vector;
    minus_one_vector.resize(names_to_appearances.size(), -1);
    
    shuffle(RingVector, names_to_appearances, minus_one_vector, rs);
    shuffle(AmuletVector, names_to_appearances, minus_one_vector, rs);
    shuffle(PotionVector, names_to_appearances, minus_one_vector, rs);
    shuffle(ScrollVector, names_to_appearances, minus_one_vector, rs);
    shuffle(SpellbookVector, names_to_appearances, minus_one_vector, rs);
    shuffle(WandVector, names_to_appearances, minus_one_vector, rs);
    
    cout << "RINGS:" << endl;
    showAppearances(RingVector, names_to_appearances);
    cout << "AMULETS:" << endl;
    showAppearances(AmuletVector, names_to_appearances);
    cout << "POTIONS:" << endl;
    showAppearances(PotionVector, names_to_appearances);
    cout << "SCROLLS:" << endl;
    showAppearances(ScrollVector, names_to_appearances);
    cout << "SPELLBOOKS:" << endl;
    showAppearances(SpellbookVector, names_to_appearances);
    cout << "WANDS:" << endl;
    showAppearances(WandVector, names_to_appearances);
}

class ConfigurationItem
{
    public:
        vector<Object>* ObjectVector;
        unsigned short NameIndex;
        unsigned short AppearanceIndex;
        
        ConfigurationItem(vector<Object>* o, unsigned short n, unsigned short a) : 
        ObjectVector(o), NameIndex(n), AppearanceIndex(a) { };
};

int readConfigurationFile(const string &conf_filename, vector<ConfigurationItem> &ci)
{
    fstream conf_stream(conf_filename.c_str(), fstream::in);
    if (conf_stream.fail())
    {
        cerr << "Could not open " << conf_filename << "." << endl;
        return -1;
    };
    
    // Type to vector
    map<string, vector<Object>* > type_to_vector;
    type_to_vector["spellbook"] = &SpellbookVector;
    type_to_vector["ring"] = &RingVector;
    type_to_vector["scroll"] = &ScrollVector;
    type_to_vector["potion"] = &PotionVector;
    type_to_vector["amulet"] = &AmuletVector;
    type_to_vector["wand"] = &WandVector;
    
    // Configuration file format is:
    // itemtype "itemname" "itemappearance"
    // e.g. spellbook "force bolt" "red"
    
    while(!conf_stream.eof())
    {
        vector<Object>* relevant_vector;
        
        string type;
        getline(conf_stream, type, ' ');
        if (conf_stream.eof())
            break;
        
        map<string, vector<Object>* >::iterator i1;
        for (i1 = type_to_vector.begin(); i1 != type_to_vector.end(); i1++)
            if (type == i1->first)
                break;
        if (i1 == type_to_vector.end())
        {
            cerr << "Unknown object type \"" << type << "\"." << endl;
            return -1;
        };
        
        relevant_vector = i1->second;
        
        int c = conf_stream.get();
        if (c != '\"')
        {
            cerr << "Expected \" after " << type << "." << endl;
            return -1;
        };
        
        string name;
        getline(conf_stream, name, '\"');
        if (conf_stream.eof())
        {
            cerr << "Unexpected end in configuration file." << endl;
            return -1;
        };
        
        int n_index = getVectorNameIndex(*relevant_vector, name);
        if (n_index == -1)
        {
            cerr << "No such " << type << ", \"" << name << "." << endl;
            return -1;
        };
        
        string appearance;
        getline(conf_stream, appearance, '\"'); // the space between name and appearance
        getline(conf_stream, appearance, '\"');
        if (conf_stream.eof())
        {
            cerr << "Unexpected end in configuration file." << endl;
            return -1;
        };
        
        int index = getVectorAppearanceIndex(*relevant_vector, appearance);
        if (index == -1)
        {
            cerr << "No such appearance in " << type << ", \"" << appearance << "." << endl;
            return -1;
        };
        
        string dummy;
        // newline
        getline(conf_stream, dummy, '\n');
        
        ci.push_back(ConfigurationItem(relevant_vector, n_index, index));
    };
    
    return 0;
};

int main(int argc, char* argv[])
{
    buildAllIndices();
    buildAllObjectVectors();
    
    string conf_filename("seed_by_discoveries.conf");
    string output_filename("seed_by_discoveries.output");
    
    unsigned int min_rand = 6, max_rand = 6;
    unsigned int start_index = 0, end_index = 1000000;
    
    unsigned int u_argc = (unsigned int) argc;

    unsigned int i1;
    for (i1 = 1; i1 < u_argc; i1++)
    {
        string argument = argv[i1];
        
        if (argument == string("-i") || argument == string("--inspect"))
        {
            if (i1 >= u_argc-2)
            {
                cerr << "Expected seed and index after -i or --inspect." << endl;
                return -1;
            }
            int seed = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            int offset = (unsigned int) strtoul(argv[i1+2], NULL, 10);
            showAllAppearances(seed, offset);
            return 0;
        }
        else if (argument == string("-c") || argument == string("--conf"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected file name after -c or --conf." << endl;
                return -1;
            };
            conf_filename = string(argv[i1+1]);
            i1++;
        }
        else if (argument == string("-o") || argument == string("--output"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected file name after -o or --output." << endl;
                return -1;
            };
            output_filename = string(argv[i1+1]);
            i1++;
        }
        else if (argument == string("--randmin"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected value after --randmin" << endl;
                return -1;
            };
            min_rand = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            i1++;
        }
        else if (argument == string("--randmax"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected value after --randmax" << endl;
                return -1;
            };
            max_rand = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            i1++;
        }
        else if (argument == string("--indexstart"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected value after --indexstart" << endl;
                return -1;
            };
            start_index = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            i1++;
        }
        else if (argument == string("--indexend"))
        {
            if (i1 == u_argc-1)
            {
                cerr << "Expected value after --indexend" << endl;
                return -1;
            };
            end_index = (unsigned int) strtoul(argv[i1+1], NULL, 10);
            i1++;
        }
        else if (argument == string("-h") || argument == string("--help"))
        {
            cout << "seed_by_discoveries --conf filename" << endl;
            cout << "seed_by_discoveries -c filename" << endl;
            cout << "    Use your own configuration file. Defaults to seed_by_discoveries.conf." << endl;
            cout << "seed_by_discoveries --output filename" << endl;
            cout << "seed_by_discoveries -o filename" << endl;
            cout << "    Output candidates in a file. Defaults to seed_by_discoveries.output." << endl;
            cout << "seed_by_discoveries --randmin value" << endl;
            cout << "    Set the minimum value of random() calls before shuffling (default 6)." << endl;
            cout << "seed_by_discoveries --randmax value" << endl;
            cout << "    Set the minimum value of random() calls before shuffling (default 6)." << endl;
            cout << "seed_by_discoveries --indexstart value" << endl;
            cout << "    Start seed forcing from this index (default 0)." << endl;
            cout << "seed_by_discoveries --indexend value" << endl;
            cout << "    Start seed forcing from this index (default 0)." << endl;
            cout << "seed_by_discoveries --help" << endl;
            cout << "seed_by_discoveries -h" << endl;
            cout << "    See this help" << endl;
            cout << "seed_by_discoveries --inspect seed random_offset" << endl;
            cout << "seed_by_discoveries -i seed random_offset" << endl;
            cout << "    Shows what objects are at given seed and pre-shuffle random() calls." << endl;
            cout << "Usage example:" << endl;
            cout << "seed_by_discoveries -c wizard_game.conf -o wizard_game.candidates --randmax 7 --indexend 1000000000" << endl;
            cout << endl;
            return 0;
        }
        else
        {
            cerr << "Unknown argument \"" << argv[i1] << "\"." << endl;
            cerr << "Use --help or -h to see accepted arguments." << endl;
            return -1;
        }
    };
    
    vector<ConfigurationItem> configuration_items;
    
    int status = readConfigurationFile(conf_filename, configuration_items);
    if (status)
        return status;
    
    fstream output_stream(output_filename.c_str(), fstream::out);
    if (output_stream.fail())
    {
        cerr << "Failed to open " << output_filename << " for writing." << endl;
        return -1;
    };
    
    vector<ConfigurationItem>::const_iterator i2;
    unsigned int index;
    
    bool bailout = false;
    
    unsigned int cur_rand;
    int start_time0 = time(0);
    int last_time0 = start_time0;

    vector<short> names_to_appearances;
    vector<short> target_mapping;
    names_to_appearances.resize(NamesToAppearances.size());
    target_mapping.resize(NameIndices.size());

    names_to_appearances = NamesToAppearances;
    
    // 0 = there are no targets
    // 1 = there is at least one ring in targets
    // 2 = at least one amulet
    // 3 = potion
    // 4 = scroll
    // 5 = spellbook
    // 6 = wand
    int target_level = 0;
    for (i2 = configuration_items.begin(); i2 != configuration_items.end(); i2++)
    {
        int new_target_level = 0;
        
        if (i2->NameIndex >= RingVector.front().NameIndex && i2->NameIndex <= RingVector.back().NameIndex)
            new_target_level = 1;
        if (i2->NameIndex >= AmuletVector.front().NameIndex && i2->NameIndex <= AmuletVector.back().NameIndex)
            new_target_level = 2;
        if (i2->NameIndex >= PotionVector.front().NameIndex && i2->NameIndex <= PotionVector.back().NameIndex)
            new_target_level = 3;
        if (i2->NameIndex >= ScrollVector.front().NameIndex && i2->NameIndex <= ScrollVector.back().NameIndex)
            new_target_level = 4;
        if (i2->NameIndex >= SpellbookVector.front().NameIndex && i2->NameIndex <= SpellbookVector.back().NameIndex)
            new_target_level = 5;
        if (i2->NameIndex >= WandVector.front().NameIndex && i2->NameIndex <= WandVector.back().NameIndex)
            new_target_level = 6;
        
        if (new_target_level > target_level)
            target_level = new_target_level;
    }
    
    if (target_level == 0)
    {
        cout << "Every seed is a candidate." << endl;
        return 0;
    }
    
    RandomState rs, return_rs;
    for (index = start_index; index <= end_index && !bailout; index++)
    {
        // Overflow protection
        if (index == 0xffffffffu)
            bailout = true;
        
	for (i1 = 0; i1 < NameIndices.size(); ++i1)
            target_mapping[i1] = -1;
        for (i2 = configuration_items.begin(); i2 != configuration_items.end(); i2++)
            target_mapping[i2->NameIndex] = i2->AppearanceIndex;
        
        return_rs.srandom(index);
        for (cur_rand = min_rand; cur_rand <= max_rand; cur_rand++)
        {
            rs = return_rs;
            for (i1 = 0; i1 < cur_rand; i1++)
                rs.random();
                
            bool result;
            result = shuffle(RingVector, names_to_appearances, target_mapping, rs);
            fixNamesToAppearances(names_to_appearances, RingVector);
            if (!result)
                continue;
            
            if (target_level <= 1)
                break;

            result = shuffle(AmuletVector, names_to_appearances, target_mapping, rs);
            fixNamesToAppearances(names_to_appearances, AmuletVector);
            if (!result)
                continue;

            if (target_level <= 2)
                break;
                
            result = shuffle(PotionVector, names_to_appearances, target_mapping, rs);
            fixNamesToAppearances(names_to_appearances, PotionVector);
            if (!result)
                continue;

            if (target_level <= 3)
                break;
            
            result = shuffle(ScrollVector, names_to_appearances, target_mapping, rs);
            fixNamesToAppearances(names_to_appearances, ScrollVector);
            if (!result)
                continue;

            if (target_level <= 4)
                break;
            
            result = shuffle(SpellbookVector, names_to_appearances, target_mapping, rs);
            fixNamesToAppearances(names_to_appearances, SpellbookVector);
            if (!result)
                continue;

            if (target_level <= 5)
                break;
                
            result = shuffle(WandVector, names_to_appearances, target_mapping, rs);
            fixNamesToAppearances(names_to_appearances, WandVector);
            if (!result)
                continue;

            break;
        };
        
        if (cur_rand <= max_rand)
        {
            cout << "New candidate: " << index << ", random offset " << cur_rand << endl;
            output_stream << index << " " << cur_rand << endl;
            output_stream.flush();
        }
        
        if (!(index % 10000))
            if (last_time0 != time(0))
            {
                last_time0 = time(0);
                
                unsigned int eta = last_time0 - start_time0;
                unsigned int finished_for_now = index - start_index;
                eta = (unsigned int) (((double) (end_index - index + 1) / (double) (finished_for_now)) * (double) eta);
                cout << "At seed " << index << ", estimated time " << eta << " seconds." << endl;
            };
    };

    unsigned int num_seeds = (end_index - start_index + 1);
    time_t seconds = time(0) - start_time0;
    cout << num_seeds << " seeds in " << seconds << " seconds." << endl;
    output_stream.close();
    
    return 0;
};

